# -*- coding: utf-8 -*-
# @老师微|信:samequant
# @网站:打板哥网 www.dabange.com
# @更多源码下载地址: https://dabange.com/download
# @有偿服务：量化课程、量化数据、策略代写、实盘对接...

from multiprocessing import Pool, cpu_count  # 并行计算
from tqdm import tqdm
from decimal import Decimal, ROUND_HALF_UP
import pandas as pd
import os
import time
from get_opening_calendar_from_szse.get_opening_calendar_from_szse import *

# 设置Pandas显示选项，保留3位小数
pd.set_option('expand_frame_repr', False)  # 当列太多时不换行
pd.set_option('display.max_rows', 6000)  # 最多显示数据的行数

class Common_functions():
    def __init__(self):
        pass

    # 导入某文件夹下所有文件的文件名
    @staticmethod
    def get_file_list_in_one_dir(path, filetype='.csv'):
        """
        从指定文件夹下，导入所有csv文件的文件名
        :param path: 文件目录
        :param filetype: 文件类型,默认.csv 比如 .csv , .xlsx
        :return: 返回所有文件名列表
        """
        file_list = []

        # 系统自带函数os.walk，用于遍历文件夹中的所有文件
        for root, dirs, files in os.walk(path):
            if files:  # 当files不为空的时候
                for f in files:
                    if f.endswith(filetype):
                        file_list.append(f[:8])

        return sorted(file_list)


class Samequant_statistic():
    def __init__(self, kline_data_path_dict=None, statistic_parameters_dict=None):
        self.stock_datas_merge_path = kline_data_path_dict['A股日k线历史行情数据所在目录']
        self.statistic_name = statistic_parameters_dict['功能名称']
        # 设置信号方法名
        self.statistic_func = self.zdt_cal_one_stock

        self.n_days = 10

        # 获取最近的交易日期
        self.g_s_o_1 = Get_stock_opening_calendar()
        self.recent_trade_date = self.g_s_o_1.get_recent_trade_date()
        self.last_trade_date = self.g_s_o_1.get_trade_date_forward_n(date_start=self.recent_trade_date, n=1)
        self.start_date = statistic_parameters_dict['开始日期']
        self.end_date = statistic_parameters_dict['结束日期']
        if self.end_date == None:
            self.end_date = self.recent_trade_date

        # 获取该class类所在.py文件的绝对目录
        self.file_full_dir = os.path.dirname(os.path.abspath(__file__))
        self.statistics_dir = self.file_full_dir + '/statistics_results'
        if not os.path.exists(self.statistics_dir):
            os.mkdir(self.statistics_dir)

        # 设置所有个股的策略信号存储路径
        self.all_statistic_path = None

    # 计算信号方法(注意该方法不能放在__main__内，否则无法用于并行计算)：反包昨日大阴线
    def zdt_cal_one_stock(self, stock_code=None):
        """
        涨停双响炮定义: 前日首板 + 昨日非涨停收阴线 + 当日触板
        """
        path = self.stock_datas_merge_path + '/{}.csv'.format(stock_code)
        df = pd.read_csv(path, parse_dates=['交易日期'], dtype={'收盘价': float, '前收盘价': float})
        date_condition = (df['交易日期'] >= (
                pd.to_datetime(self.start_date) - pd.to_timedelta(self.n_days + 10, unit='D'))) & (
                                 df['交易日期'] <= pd.to_datetime(self.end_date))
        df = df[date_condition]

        if df.empty:
            return pd.DataFrame()

        # 去重复值
        df.drop_duplicates(subset=['交易日期'], keep='last', inplace=True)
        df['上市天数'] = df.index + 1
        df = df[(df['交易日期'] >= pd.to_datetime('19961226')) & (df['上市天数'] > 20)]
        if df.empty:
            return pd.DataFrame()
        df.reset_index(drop=True, inplace=True)
        df = self.cal_fuquan_price(df=df, fuquan_type='前复权')
        df = df.round(2)

        # 计算涨跌停价
        df = self.cal_zdt_price(df=df)

        df['次日涨幅'] = df['涨跌幅'].shift(-1)
        df['10日涨跌幅'] = df['收盘价_复权'].shift(-10)/df['收盘价_复权'] - 1

        self.actor_zhangting_num = '近{}个交易日涨停次数'.format(self.n_days)
        df[self.actor_zhangting_num] = df['收盘涨停'].shift(1).rolling(self.n_days).sum()

        df[self.statistic_name] = None
        df.loc[(df[self.actor_zhangting_num] <= 5) & (df['收盘涨停'].shift(1) < 1) & (
                df['收盘涨停'] == 1), '首板'] = 1
        c1 = df['首板'].shift(2) == 1
        c21 = (df['收盘价_复权'].shift(1) < df['开盘价_复权'].shift(1)) & (df['最高价_复权'].shift(1)/df['收盘价_复权'].shift(1) > 1.035)
        c22 = df['涨跌幅'].shift(1) <= -3.5
        c3 = df['涨停触板'] == 1

        df[self.statistic_name] = 0
        df.loc[c1 & (c21 | c22) & c3, self.statistic_name] = 1

        df = df[df[self.statistic_name] == 1]

        df[f'{self.statistic_name}封板率'] = None
        df[f'{self.statistic_name}触板次数'] = None
        # 给最后一行的列'首板触板封板率'赋值
        if len(df) > 0:
            df.iloc[-1, df.columns.get_loc(f'{self.statistic_name}触板次数')] = len(df)
            df.iloc[-1, df.columns.get_loc(f'{self.statistic_name}封板率')] = len(df[df['收盘涨停'] == 1]) / len(df)

        df.reset_index(inplace=True, drop=True)

        if not df.empty:
            # print(df.columns)
            df = df[['交易日期', '股票代码', '股票名称', '收盘价', '最高价', '涨停价', '最低价', '开盘价', '前收盘价', '涨跌额',
                     '涨跌幅', '换手率', '成交量', '总市值', '流通市值', '成交额', '次日涨幅', '10日涨跌幅', '涨停触板', '一字涨停', '收盘涨停', '涨停炸板',
                     f'{self.statistic_name}触板次数', f'{self.statistic_name}封板率']]
            df.reset_index(drop=True, inplace=True)
        else:
            df = pd.DataFrame()
        return df

    def cal_zdt_price(self, df):
        """
        计算股票当天的涨跌停价格。在计算涨跌停价格的时候，按照严格的四舍五入。
        包含st股，但是不包含新股
        涨跌停制度规则:
            ---2020年8月3日
            非ST股票 10%
            ST股票 5%

            ---2020年8月4日至今
            普通非ST股票 10%
            普通ST股票 5%

            科创板（sh68） 20%
            创业板（sz30） 20%
            科创板和创业板即使ST，涨跌幅限制也是20%

            北交所（bj） 30%

        :param df: 必须得是日线数据。必须包含的字段：前收盘价，开盘价，最高价，最低价
        :return:
        """
        # 计算涨停价格
        # 普通股票
        cond_st = (df['股票名称'].str.contains('ST') | df['股票名称'].str.contains('st'))
        df['涨停价'] = df['前收盘价'] * 1.1
        df['跌停价'] = df['前收盘价'] * 0.9
        df.loc[cond_st, '涨停价'] = df['前收盘价'] * 1.05
        df.loc[cond_st, '跌停价'] = df['前收盘价'] * 0.95

        # 2020年8月23日之后涨跌停规则有所改变
        # 新规的科创板
        new_rule_kcb = (df['交易日期'] > pd.to_datetime('2020-08-23')) & df['股票代码'].str.contains('sh68')
        # 新规的创业板
        new_rule_cyb = (df['交易日期'] > pd.to_datetime('2020-08-23')) & df['股票代码'].str.contains('sz30')
        # 北交所条件
        cond_bj = df['股票代码'].str.contains('bj')

        # 科创板 & 创业板
        df.loc[new_rule_kcb | new_rule_cyb, '涨停价'] = df['前收盘价'] * 1.2
        df.loc[new_rule_kcb | new_rule_cyb, '跌停价'] = df['前收盘价'] * 0.8

        # 北交所
        df.loc[cond_bj, '涨停价'] = df['前收盘价'] * 1.3
        df.loc[cond_bj, '跌停价'] = df['前收盘价'] * 0.7

        # 四舍五入
        df['涨停价'] = df['涨停价'].apply(
            lambda x: float(Decimal(x + 1e-7).quantize(Decimal('1.00'), rounding=ROUND_HALF_UP)))
        df['跌停价'] = df['跌停价'].apply(
            lambda x: float(Decimal(x + 1e-7).quantize(Decimal('1.00'), rounding=ROUND_HALF_UP)))

        # 判断是否一字涨停
        df['一字涨停'] = 0
        df.loc[df['最低价'] == df['涨停价'], '一字涨停'] = 1
        # 判断是否一字跌停
        df['一字跌停'] = 0
        df.loc[df['最高价'] == df['跌停价'], '一字跌停'] = 1
        # 判断是否开盘涨停
        df['开盘涨停'] = 0
        df.loc[df['开盘价'] == df['涨停价'], '开盘涨停'] = 1
        # 判断是否开盘跌停
        df['开盘跌停'] = 0
        df.loc[df['开盘价'] == df['跌停价'], '开盘跌停'] = 1
        # 判断是否涨停收盘，涨停不能买
        df['收盘涨停'] = 0
        df.loc[df['收盘价'] == df['涨停价'], '收盘涨停'] = 1
        # 判断是否收盘跌停，跌停可买，跌停不卖
        df['收盘跌停'] = 0
        df.loc[df['收盘价'] == df['跌停价'], '收盘跌停'] = 1

        df['涨停触板'] = 0
        df.loc[df['最高价'] == df['涨停价'], '涨停触板'] = 1

        # 判断涨停炸板
        df['涨停炸板'] = 0
        df.loc[(df['最高价'] == df['涨停价']) & (df['收盘价'] < df['涨停价']), '涨停炸板'] = 1

        # 非一字板涨停
        df['非一涨停'] = 0
        df.loc[(df['开盘涨停'] == 0) & (df['收盘涨停'] == 1), '非一涨停'] = 1
        # 非一字板炸板
        df['非一炸板'] = 0
        df.loc[(df['开盘涨停'] == 0) & (df['涨停炸板'] == 1), '非一炸板'] = 1
        # 非一字板跌停
        df['非一跌停'] = 0
        df.loc[(df['开盘跌停'] == 0) & (df['收盘跌停'] == 1), '非一跌停'] = 1

        return df

    def cal_fuquan_price(self, df, fuquan_type='前复权'):
        """
        用于计算复权价格
        :param df: 必须包含的字段：收盘价，前收盘价，开盘价，最高价，最低价
        :param fuquan_type: ‘前复权’或者‘后复权’
        :return: 最终输出的df中，新增字段：收盘价_复权，开盘价_复权，最高价_复权，最低价_复权
        """

        # 计算复权因子：假设你一开始有1元钱，投资到这个股票，最终会变成多少钱.
        # 注:不能用 df['复权因子'] = (df['收盘价'] / df['前收盘价'] - 1).cumprod() 因为上市首日有的数据里前收盘价为0
        df['复权因子'] = (1 + df['涨跌幅'] / 100).cumprod()

        # 计算前复权、后复权收盘价
        if fuquan_type == '后复权':
            df['收盘价_复权'] = df['复权因子'] * (df.iloc[0]['收盘价'] / df.iloc[0]['复权因子'])
        elif fuquan_type == '前复权':
            df['收盘价_复权'] = df['复权因子'] * (df.iloc[-1]['收盘价'] / df.iloc[-1]['复权因子'])
        else:
            raise ValueError('计算复权价时，出现未知的复权类型：%s' % fuquan_type)

        # 计算复权
        df['开盘价_复权'] = df['开盘价'] / df['收盘价'] * df['收盘价_复权']
        df['最高价_复权'] = df['最高价'] / df['收盘价'] * df['收盘价_复权']
        df['最低价_复权'] = df['最低价'] / df['收盘价'] * df['收盘价_复权']
        del df['复权因子']
        return df

    # 定义一个高阶函数，接受一个函数和该函数所需的参数
    def cal_one_stock_statistic(self, signal_func=None, *args, **kwargs):
        if signal_func == None:
            df = self.signal_func(*args, **kwargs)
        else:
            df = signal_func(*args, **kwargs)
        return df

    def cal_all_statistic_csv(self):
        # 设置所有个股的策略信号存储路径
        self.all_statistic_path = self.statistics_dir + f'/all_statistics_{self.start_date}_{self.end_date}_{self.statistic_name}.csv'

        if os.path.exists(self.all_statistic_path):
            return

        # 读取本机所有A股历史行情数据目录下的.csv文件返回股票列表
        stock_list0 = Common_functions.get_file_list_in_one_dir(path=self.stock_datas_merge_path, filetype='.csv')

        stock_list = sorted(list(set(stock_list0)))

        # 并行运行计算所有股票信号
        start_time = time.time()  # 标记开始时间
        multiply_process = True  # 设置是否并行运行
        if multiply_process:
            # 给pool并行传入2个以上的参数，先转为列表，每个元素是一个元组。对应self.cal_one_stock_statistic
            arguments = []
            for stock_code in stock_list:
                arguments.append((self.statistic_func, stock_code))

            print(f'开始并行运行计算所有股票{self.statistic_name}信号...')
            with Pool(max(cpu_count() - 1, 1)) as pool:
                # 使用starmap方法（传入单个参数用pool.map）并行运行 cal_one_stock_statistic，并传入参数列表使用并行批量获得dataframe的一个列表
                df_list = pool.starmap(self.cal_one_stock_statistic, arguments)
            print('并行计算完成。耗时：', time.time() - start_time)
        else:
            df_list = []
            for stock_code in stock_list:
                df_one = self.cal_one_stock_statistic(statistic_func=self.statistic_func, stock_code=stock_code)
                df_list.append(df_one)
            print('for循环计算完成。耗时：', time.time() - start_time)
        # 将所有个股的信号df_list合并为一个总表
        df = pd.concat(df_list, ignore_index=True)

        df = df.sort_values(by=['交易日期'])
        df = df[(pd.to_datetime(df['交易日期']) >= pd.to_datetime(self.start_date)) & (
                pd.to_datetime(df['交易日期']) <= pd.to_datetime(self.end_date))]

        df.to_csv(self.all_statistic_path, mode='w', index=False)
        print('计算所有数据完成！')
        print(f'***统计期间 {self.statistic_name}-总数为{len(df)}个***')

    def statistics(self, only_10cm=True):
        print('涨停双响炮定义: 前日首板涨停 + 昨日非涨停（跌幅超-3.5%或长上影阴线） + 当日触板 + 上市大于20个交易日。\n中国A股现行的涨跌停板制度是1996年12月26日开始实施的。 ')

        # 设置所有个股的策略信号存储路径
        self.all_statistic_path = self.statistics_dir + f'/all_statistics_{self.start_date}_{self.end_date}_{self.statistic_name}.csv'

        df_s = pd.DataFrame()
        if os.path.exists(self.all_statistic_path):
            df = pd.read_csv(self.all_statistic_path)
            if only_10cm:
                print(f'总数为:{len(df)}')
                df = df[~(df['股票名称'].str.contains('ST') | df['股票代码'].str.startswith('sh68') | (
                        df['股票代码'].str.startswith('sz3') & (
                        pd.to_datetime(df['交易日期']) > pd.to_datetime('2020-08-23'))))]
                print(f'过滤st、科创板、20cm创业板后的总数为:{len(df)}')

            df.loc[df['涨停炸板'] == 1, '炸板当日浮亏比例'] = df['收盘价'] / df['涨停价'] - 1
            # print(df.tail(100))
            # exit()
            total = len(df)
            sbzt_num = len(df[df['收盘涨停'] == 1])
            sbyzzt_num = len(df[df['一字涨停'] == 1])
            sbzb_num = len(df[df['涨停炸板'] == 1])
            sbzt_sz_num = len(df[(df['收盘涨停'] == 1) & (df['次日涨幅'] > 0)])
            sbyzzt_sz_num = len(df[(df['一字涨停'] == 1) & (df['次日涨幅'] > 0)])
            sbzb_sz_num = len(df[(df['涨停炸板'] == 1) & (df['次日涨幅'] > 0)])

            sbzt_avg_pct = df[df['收盘涨停'] == 1]['次日涨幅'].mean()
            sbzt_not_yzb_num = len(df[(df['收盘涨停'] == 1) & (df['一字涨停'] < 1)])
            sbzt_not_yzb_avg_pct = df[(df['收盘涨停'] == 1) & (df['一字涨停'] < 1)]['次日涨幅'].mean()

            df_s.loc[0, f'{self.statistic_name}触板总数'] = total
            df_s.loc[0, f'{self.statistic_name}涨停总数'] = sbzt_num
            df_s.loc[0, f'{self.statistic_name}炸板总数'] = sbzb_num
            df_s.loc[0, f'{self.statistic_name}非一涨停总数'] = sbzt_not_yzb_num
            df_s.loc[0, f'{self.statistic_name}封板率'] = sbzt_num / total
            df_s.loc[0, f'{self.statistic_name}非一封板率'] = sbzt_not_yzb_num / (total - sbyzzt_num)
            df_s.loc[0, f'{self.statistic_name}涨停次日收涨概率'] = '{:.2%}'.format(sbzt_sz_num / sbzt_num)
            df_s.loc[0, f'{self.statistic_name}一字涨停次日收涨概率'] = '{:.2%}'.format(sbyzzt_sz_num / sbyzzt_num)
            df_s.loc[0, f'{self.statistic_name}炸板次日收涨概率'] = '{:.2%}'.format(sbzb_sz_num / sbzb_num)

            sbzb_ks_pct = df[df['涨停炸板'] == 1]['炸板当日浮亏比例'].mean()
            sbzb_avg_pct = df[df['涨停炸板'] == 1]['次日涨幅'].mean()

            df_s.loc[0, f'{self.statistic_name}涨停次日平均涨幅'] = sbzt_avg_pct
            df_s.loc[0, f'{self.statistic_name}非一涨停次日平均涨幅'] = sbzt_not_yzb_avg_pct
            df_s.loc[0, f'{self.statistic_name}炸板当日平均浮亏比例'] = sbzb_ks_pct
            df_s.loc[0, f'{self.statistic_name}炸板次日平均涨幅'] = sbzb_avg_pct

            df_s.loc[0, f'{self.statistic_name}预估净值'] = df_s.loc[0, f'{self.statistic_name}非一封板率'] * (1 + df_s.loc[0, f'{self.statistic_name}非一涨停次日平均涨幅'] / 100) + (
                    1 - df_s.loc[0, f'{self.statistic_name}封板率']) * (
                                            1 + (df_s.loc[0, f'{self.statistic_name}炸板当日平均浮亏比例'] + df_s.loc[0, f'{self.statistic_name}炸板次日平均涨幅']) / 100)
            df_s.loc[0, f'{self.statistic_name}封板率'] = '{:.2%}'.format(sbzt_num / total)
            df_s.loc[0, f'{self.statistic_name}非一封板率'] = '{:.2%}'.format(sbzt_not_yzb_num / (total - sbyzzt_num))
            df_s.loc[0, f'{self.statistic_name}涨停次日平均涨幅'] = '{:.2%}'.format(sbzt_avg_pct / 100)
            df_s.loc[0, f'{self.statistic_name}非一涨停次日平均涨幅'] = '{:.2%}'.format(sbzt_not_yzb_avg_pct / 100)
            df_s.loc[0, f'{self.statistic_name}炸板当日平均浮亏比例'] = '{:.2%}'.format(sbzb_ks_pct)
            df_s.loc[0, f'{self.statistic_name}炸板次日平均涨幅'] = '{:.2%}'.format(sbzb_avg_pct / 100)
            print(df_s.T)
            # exit()
            df = df[df[f'{self.statistic_name}触板次数'] > 0]
            df_head100 = df.sort_values(by=[f'{self.statistic_name}封板率', f'{self.statistic_name}触板次数'], ascending=[False, False]).head(10)
            df_head100.reset_index(drop=True, inplace=True)
            df_tail100 = df.sort_values(by=[f'{self.statistic_name}封板率', f'{self.statistic_name}触板次数'], ascending=[True, False]).head(10)
            df_tail100.reset_index(drop=True, inplace=True)
            print('\n')
            print(f'{self.statistic_name}触板封板率前10榜单：')
            print(df_head100[['交易日期', '股票代码', '股票名称', f'{self.statistic_name}触板次数', f'{self.statistic_name}封板率', '炸板当日浮亏比例']])
            print('\n')
            print(f'{self.statistic_name}触板封板率倒数前10榜单：')
            print(df_tail100[['交易日期', '股票代码', '股票名称', f'{self.statistic_name}触板次数', f'{self.statistic_name}封板率', '炸板当日浮亏比例']])

            df_pct10 = df.sort_values(by='10日涨跌幅', ascending=False).head(10)
            # print(df_pct10)
            df_pct10.reset_index(drop=True, inplace=True)
            print(f'{self.statistic_name}10日涨幅top前10榜单：')
            print(df_pct10[['交易日期', '股票代码', '股票名称', '10日涨跌幅']])
            df_pct_tail10 = df.sort_values(by='10日涨跌幅', ascending=True).head(10)
            # print(df_pct10)
            df_pct_tail10.reset_index(drop=True, inplace=True)
            print(f'{self.statistic_name}10日涨幅倒数前10榜单：')
            print(df_pct_tail10[['交易日期', '股票代码', '股票名称', '10日涨跌幅']])


if __name__ == '__main__':
    """
    据说涨停双响炮容易出牛股，是真的吗？每期量化一个特征，本期统计涨停双响炮。各位学员好，这里是samequant量化，
    只讲干货，适合理性的人观看。

    统计结果包括涨停双响炮的封板率、封板次日收涨概率、封板次日平均涨幅、炸板率、炸板次日平均涨幅、炸板次日收涨概率、
    排除一字板后的封板率、双响炮当日封板率top排行榜、双响炮后10日收益率排行榜等。

    我们运行程序看下统计的结果，然后详细介绍。
    
    好了，本期的分享就到这里，如果对量化统计感兴趣的朋友，欢迎点赞关注。
    
    双响炮定义: 前日首板 + 昨日非涨停高开低走收大阴线 + 当日触板 + 上市大于20个交易日
    中国A股现行的涨跌停板制度是1996年12月26日开始实施的。
    """
    kline_data_path_dict = {
        'A股日k线历史行情数据所在目录': r'C:\fastapi_demo\datas\stock_datas_merge'
    }

    statistic_parameters_dict = {'功能名称': '涨停双响炮',
                                 '开始日期': '19961226',
                                 '结束日期': None
                                 }
    s_s_1 = Samequant_statistic(kline_data_path_dict=kline_data_path_dict,
                                statistic_parameters_dict=statistic_parameters_dict)
    # 计算单只个股的双响炮 sz002016
    df_one = s_s_1.zdt_cal_one_stock(stock_code='sh600455')
    print(df_one)
    exit()
    # 并行计算全量个股首板数据
    s_s_1.cal_all_statistic_csv()

    # 全量首板统计分析
    s_s_1.statistics()

    # 10%涨跌幅股票池统计分析
    # s_s_1.statistics(only_10cm=True)

    """
    SameQuant持续的分享交易统计分析数据，欢迎点赞关注。
    """
